#include <cmath>
#include <iostream>
#include <vector>
using namespace std;

// One variable function 'f(x)'
class scalarfun1d_t {   //L-scfun1d-b
public:
  virtual double eval(double x)=0;
};
//L-scfun1d-e

// Two variables function 'f(x,y)'
class scalarfun2d_t {
public:
  virtual double eval(double x,double y)=0;
};

// Two variables function 'f(x,y,z)'
class scalarfun3d_t {
public:
  virtual double eval(double x,double y,double z)=0;
};

// Integral of a one var funct over an interval
class integral1d_t { //L-int1d-b
private:
  int N;                        // Nbr of integration pts
  double xa,xb;                 // integration interval
  scalarfun1d_t *fp;              // ptr to function
public:
  // Ctor, set values
  integral1d_t(int Na=0,double xaa=NAN,
               double xba=NAN,scalarfun1d_t *fpa=NULL) {
    set(Na,xaa,xba,fpa);
  }
  // Set values
  void set(int Na,double xaa,double xba,scalarfun1d_t *fpa) {
    N=Na; xa=xaa; xb=xba; fp=fpa;
  }
  // Computes the integral using Simpson's rule
  double integral() {  
    double
      h = (xb-xa)/N,
      h2 = h/2.0,
      finteg=0.0;
    for (int j=0; j<N; j++) {
      double x = j*h;
      double f = fp->eval(x);
      finteg += 2.0*f;
      finteg += 4.0*fp->eval(x+h2);
    }
    finteg -= fp->eval(xa);
    finteg += fp->eval(xb);
    finteg *= h/6.0;
    return finteg;
  }
};
//L-int1d-e

// Integral of 2var functions over a rectangle
class integral2d_t { //L-int2dok-b
private:
  scalarfun2d_t *f2dp;          // Ptr to 2var funct
  integral1d_t int1dy, int1dx;  // integrals over x and y
  // For a given y gives the function fy(x) = f(x,y)
  class fy_t : public scalarfun1d_t {
    double y;
    integral2d_t *i2dp;         // ptr to calling class
    friend class integral2d_t;
  public:
    double eval(double x) { return i2dp->f2dp->eval(x,y); }
  } fy;
  // For a given y, integral over x: g(y) = int_{x=xa}^xb f(x,y) dx
  class g_t : public scalarfun1d_t {
    integral2d_t *i2dp;         // ptr to calling class
    friend class integral2d_t;
  public:
    double eval(double y) { // wrapper pass to calling class
      return i2dp->geval(y);
    }
  } g;
  double geval(double y) {
    // Integrates over x for y=cnst
    fy.y = y;
    return int1dx.integral();
  }
  int N;                        // Nbr of integr segments
  double xa,xb,ya,yb;           // rectangle corners
  friend class fy_t; 
  friend class g_t;
public:
  void set(int Na=0,double xaa=NAN,double xba=NAN,
           double yaa=NAN,double yba=NAN,
           scalarfun2d_t *f2dpa=NULL) {
    N = Na; xa=xaa; xb=xba; ya=yaa; yb=yba;
    f2dp = f2dpa;
    int1dy.set(N,ya,yb,&g);
    int1dx.set(N,xa,xb,&fy);
  }
  integral2d_t(int Na=0,          // Ctor
               double xaa=NAN,double xba=NAN,
               double yaa=NAN,double yba=NAN,
               scalarfun2d_t *f2dpa=NULL) {
    set(Na,xaa,xba,yaa,yba,f2dpa);
    g.i2dp = this;
    fy.i2dp = this;
  }
  double integral() {
    // 2D integral. Integrates over y the
    // integral over x
    return int1dy.integral();
  }
};
//L-int2dok-e

// Integral of 3var functions over a cube
class integral3d_t { //L-int3d-b
private:
  scalarfun3d_t *f3dp;          // Ptr to 3var funct
  integral2d_t int2dxy;         // integral over x,y
  integral1d_t int1dz;  // integral over z
  // For a given z gives the function fz(x,y) = f(x,y,z)
  class fz_t : public scalarfun2d_t {
    double z;
    integral3d_t *i3dp;         // ptr to calling class
    friend class integral3d_t;
  public:
    double eval(double x,double y) {
      return i3dp->f3dp->eval(x,y,z);
    }
  } fz;
  // For a given z, gives g(z) = integral over x,y:
  // g(z) = int_{x=xa}^xb int_{y=ya}^yb  f(x,y,z) dx dy
  class g_t : public scalarfun1d_t {
    integral3d_t *i3dp;         // ptr to calling class
    friend class integral3d_t;
  public:
    double eval(double z) { // wrapper pass to calling class
      return i3dp->geval(z);
    }
  } g;
  double geval(double z) {
    // Integrates over x,y for z=cnst
    fz.z = z;
    return int2dxy.integral();
  }
  int N;                        // Nbr of integr segments
  double xa,xb,ya,yb,za,zb;     // rectangle corners
  friend class fz_t; 
  friend class g_t;
public:
  void set(int Na,double xaa,double xba, // Set from args
           double yaa,double yba,
           double zaa,double zba,
           scalarfun3d_t *f3dpa) {
    N = Na; xa=xaa; xb=xba; ya=yaa; yb=yba; za=zaa; zb=zba;
    f3dp = f3dpa;
    int2dxy.set(N,xa,xb,ya,yb,&fz);
    int1dz.set(N,za,zb,&g);
  }
  integral3d_t(int Na=0,          // Ctor
               double xaa=NAN,double xba=NAN,
               double yaa=NAN,double yba=NAN,
               double zaa=NAN,double zba=NAN,
               scalarfun3d_t *f3dpa=NULL) {
    set(Na,xaa,xba,yaa,yba,zaa,zba,f3dpa);
    g.i3dp = this;
    fz.i3dp = this;
  }
  double integral() { // Comp 3D integral. Integrates over z
    return int1dz.integral();
  }
};
//L-int3d-e

// User code starts here----
// Specific 1var function to be integrated
class fun_t : public scalarfun1d_t { //L-usercode-b
public:
  double eval(double x) {
    // Must give (2/pi) = 0.63662
    return sin(0.5*M_PI*x); 
  }
} fun;

// Specific 2var function to be integrated
class fun2_t : public scalarfun2d_t {
public:
  double eval(double x,double y) {
    // Must give (2/pi)^2 = 0.40528
    return sin(0.5*M_PI*x)*sin(0.5*M_PI*y); 
  }
} fun2;

// Specific 3var function to be integrated
class fun3_t : public scalarfun3d_t {
public:
  double eval(double x,double y,double z) {
    // Must give (2/pi)^3 = 0.25801
    return sin(0.5*M_PI*x)
      *sin(0.5*M_PI*y)*sin(0.5*M_PI*z); 
  }
} fun3;

int main() {
  int N=10;
  // 1D integration example
  integral1d_t int1d(N,0,1.0,&fun);
  cout << "integral(x) = " << int1d.integral() << endl;

  // 2D integration example
  integral2d_t int2d(N,0.0,1.0,0.0,1.0,&fun2);
  cout << "integral(x,y) = " << int2d.integral() << endl;

  // 3D integration example
  integral3d_t int3d(N,0.0,1.0,0.0,1.0,0.0,1.0,&fun3);
  cout << "integral(x,y,z) = " << int3d.integral() << endl;
  
  return 0;
}
//L-usercode-e
